"""Taxonomies."""


class PageType:
    CATAGORY = "cat"
    SEARCH = "search"
    TREND = "trends"
    LIST = "slideshow"


class APIAction:
    LP = "lp"


class AJAXIncrement:
    ONE = 1
    ZERO = 0
